<?php

namespace Drupal\openapi\Controller;

use Drupal\Component\Serialization\Exception\InvalidDataTypeException;
use Drupal\Core\Cache\CacheableMetadata;
use Drupal\Core\Cache\CacheableResponse;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Render\RenderContext;
use Drupal\Core\Render\RendererInterface;
use Drupal\openapi\Plugin\openapi\OpenApiGeneratorInterface;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\Yaml\Dumper;
use Symfony\Component\Yaml\Yaml as SymfonyYaml;

/**
 * API Specification controller base.
 */
class ApiSpecificationController extends ControllerBase {

  public function __construct(
    protected RendererInterface $renderer,
  ) {}

  /**
   * Gets the OpenAPI output in JSON format.
   *
   * @return string
   *   The page title.
   */
  public function title(OpenApiGeneratorInterface $generator, Request $request) {
    return $this->t(
      '%label OpenApi Schema Download',
      [
        '%label' => $generator->getLabel(),
      ]
    );
  }

  /**
   * Gets the OpenAPI output.
   *
   * @param OpenApiGeneratorInterface $openapi_generator
   *   Generator plugin.
   * @param Request $request
   *   Request object.
   * @param string $content_type
   *   Content type to return.
   *
   * @return \Symfony\Component\HttpFoundation\Response
   */
  public function getSpecification(OpenApiGeneratorInterface $openapi_generator, Request $request, string $content_type = 'json') {
    $options = $request->get('options', []);
    $openapi_generator->setOptions($options + $openapi_generator->getOptions());
    // Throw out cacheable metadata during rendering, e.g. from generating
    // examples.
    $spec = $this->renderer->executeInRenderContext(new RenderContext(), fn () => $openapi_generator->getSpecification());
    $response = $content_type === 'yaml'
      ? new CacheableResponse(
        // @todo Rip out this boilerplate if https://www.drupal.org/project/drupal/issues/3457164 goes in.
        self::encodeYaml($spec),
        headers: ['content-type' => 'text/yaml'],
      )
      : new CacheableResponse(
        json_encode($spec),
        headers: ['content-type' => 'application/json'],
      );
    $response->addCacheableDependency(
      (new CacheableMetadata())
        ->addCacheContexts(['url.site', 'url.query_args:options', 'user.permissions'])
        ->addCacheTags(['entity_types', 'entity_field_info'])
    );
    return $response;
  }

  /**
   * Customized version of YamlSymfony::encode() with additional flags.
   *
   * @param mixed $data
   *   Data to encode.
   *
   * @return string
   *   Encoded data.
   */
  protected static function encodeYaml(mixed $data) {
    try {
      // Set the indentation to 2 to match Drupal's coding standards.
      $yaml = new Dumper(2);
      // Yaml and JSON interoperability for empty arrays and quoted response
      // codes.
      $additionalFlags = SymfonyYaml::DUMP_OBJECT_AS_MAP | SymfonyYaml::DUMP_NUMERIC_KEY_AS_STRING;
      return $yaml->dump($data, PHP_INT_MAX, 0, SymfonyYaml::DUMP_EXCEPTION_ON_INVALID_TYPE | SymfonyYaml::DUMP_MULTI_LINE_LITERAL_BLOCK | $additionalFlags);
    }
    catch (\Exception $e) {
      throw new InvalidDataTypeException($e->getMessage(), $e->getCode(), $e);
    }
  }

}
